package org.atea.nlptools.ocr.services;

import java.io.File;

import org.atea.nlptools.ocr.Exceptions.LeptonicaException;
import org.atea.nlptools.ocr.Exceptions.TesseractException;
import org.atea.nlptools.ocr.abstractions.objects.IOcrOptions;
import org.atea.nlptools.ocr.abstractions.services.IOcrService;
import org.atea.nlptools.ocr.objects.TesseractOptions;

import org.bytedeco.javacpp.BytePointer;
import org.bytedeco.leptonica.PIX;
import org.bytedeco.leptonica.global.lept;
import org.bytedeco.tesseract.TessBaseAPI;

/**
 * Represents a service for performing OCR on files using the Tesseract engine.
 */
public class TesseractOcrService implements IOcrService
{
    private final String tessData;

    /**
     * Initializes a new instance of the {@link TesseractOcrService} class.
     * @param tesseractDataPath Path to the tesseract data directory.
     */
    public TesseractOcrService(String tesseractDataPath)
    {
        this.tessData = tesseractDataPath;
    }

    @Override
    public String run(File file, IOcrOptions options)
        throws Exception, IllegalArgumentException
    {
        if (options.getClass() != TesseractOptions.class)
        {
            throw new IllegalArgumentException("Expected an options parameter of type " + TesseractOptions.class);
        }

        TesseractOptions tOptions = (TesseractOptions)options;
        PIX inputImage = null;
        TessBaseAPI api = null;
        BytePointer outputTextPtr = null;

        try
        {
            inputImage = lept.pixRead(file.getAbsolutePath());
            if (inputImage == null) {
                throw new LeptonicaException("Didn't recognise the image format. Check stdout for more info.");
            }

            api = new TessBaseAPI();

            if (api.Init(this.tessData, tOptions.getLanguage()) != 0) {
                throw new TesseractException("Could not initialize tesseract. Check stdout for more info");
            }

            api.SetPageSegMode(tOptions.pageSegmentationMode);
            api.SetImage(inputImage);
            outputTextPtr = api.GetUTF8Text();

            return outputTextPtr.getString();
        }
        finally
        {
            if (inputImage != null)
            {
                lept.pixDestroy(inputImage);
            }

            if (api != null)
            {
                api.End();
                api.close();
            }

            if (outputTextPtr != null)
            {
                outputTextPtr.deallocate();
                outputTextPtr.close();
            }
        }
    }
}
